# coding=utf-8
"""Diagnostic to compute regional averages"""
import iris
import iris.util
import iris.coords
import iris.analysis
import iris.exceptions
from iris.coord_categorisation import add_categorised_coord
from iris.cube import Cube, CubeList

import numpy as np

from earthdiagnostics.box import Box
from earthdiagnostics.constants import Basins
from earthdiagnostics.diagnostic import Diagnostic, DiagnosticOption, DiagnosticIntOption, DiagnosticDomainOption, \
    DiagnosticBoolOption, DiagnosticBasinOption, DiagnosticVariableOption
from earthdiagnostics.modelingrealm import ModelingRealms
from earthdiagnostics.utils import Utils, TempFile


class ZonalMean(Diagnostic):
    """
    Computes the zonal mean value of the field (weighted).

    :param data_manager: data management object
    :type data_manager: DataManager
    :param startdate: startdate
    :type startdate: str
    :param member: member number
    :type member: int
    :param chunk: chunk's number
    :type chunk: int
    :param variable: variable to average
    :type variable: str
    :param box: box used to restrict the vertical mean
    :type box: Box
    """

    alias = 'zonmean'
    "Diagnostic alias for the configuration file"

    def __init__(self, data_manager, startdate, member, chunk, domain, variable, basin, grid_point):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.domain = domain
        self.variable = variable
        self.basin = basin
        self.grid_point = grid_point

        self.declared = {}

        self.lat_name = 'lat'
        self.lon_name = 'lon'

    def __eq__(self, other):
        if self._different_type(other):
            return False
        return self.startdate == other.startdate and self.member == other.member and self.chunk == other.chunk and \
            self.box == other.box and self.variable == other.variable

    def __str__(self):
        return 'Zonal mean Startdate: {0.startdate} Member: {0.member} Chunk: {0.chunk} Variable: {0.variable} ' \
               'Grid point: {0.grid_point}'.format(self)

    def __hash__(self):
        return hash(str(self))

    @classmethod
    def generate_jobs(cls, diags, options):
        """
        Create a job for each chunk to compute the diagnostic

        :param diags: Diagnostics manager class
        :type diags: Diags
        :param options: variable, minimum depth (level), maximum depth (level)
        :type options: list[str]
        :return:
        """
        options_available = (DiagnosticDomainOption(),
                             DiagnosticVariableOption(diags.data_manager.config.var_manager),
                             DiagnosticOption('grid_point', 'T'),
                             DiagnosticBasinOption('basin', Basins().Global),
                             DiagnosticOption('grid', ''))
        options = cls.process_options(options, options_available)

        job_list = list()
        for startdate, member, chunk in diags.config.experiment.get_chunk_list():
            job = ZonalMean(diags.data_manager, startdate, member, chunk,
                             options['domain'], options['variable'], options['basin'],
                             options['grid_point'].lower())
            job_list.append(job)

        return job_list

    def request_data(self):
        """Request data required by the diagnostic"""
        self.variable_file = self.request_chunk(
            self.domain, self.variable, self.startdate, self.member, self.chunk
        )

    def declare_data_generated(self):
        """Declare data to be generated by the diagnostic"""
        self.zonal_mean = self.declare_chunk(
            ModelingRealms.ocean, self.variable + 'zonal',
            self.startdate, self.member, self.chunk, region=self.basin
        )


    def compute(self):
        """Run the diagnostic"""



        has_levels = self._fix_file_metadata()
        data = self._load_data()
        if has_levels:
            self._meand_3d_variable(data)

    def _meand_3d_variable(self, data):
        mask = np.squeeze(Utils.get_mask(self.basin, True))
        data = data * mask
        latitude = data.coord('latitude')
        add_categorised_coord(
            data,
            'latitude_degree',  'latitude',
            lambda coord, value: round(value),
            units=data.coord('latitude').units,
        )
        data.remove_coord('latitude')
        data.remove_coord('longitude')
        mean = iris.cube.CubeList()
        for map_slice in data.slices_over('time'):
            reshaped = map_slice.data.reshape((map_slice.shape[0], map_slice.shape[1] * map_slice.shape[2]))
            reshaped = Cube(
                reshaped,
            )
            reshaped.add_dim_coord(map_slice.coord('depth'), 0)
            reshaped.add_aux_coord(map_slice.coord('time'))
            reshaped.add_aux_coord(map_slice.coord('latitude_degree').copy(map_slice.coord('latitude_degree').points.reshape(map_slice.shape[1] * map_slice.shape[2])), 1)
            reshaped = reshaped.aggregated_by(
                'latitude_degree',
                iris.analysis.MEAN,
            )
            mean.append(CubeList((c for c in reshaped.slices_over('latitude_degree'))).merge_cube())
        cube = mean.merge_cube()
        cube.var_name = 'result'
        cube.units = data.units
        temp = TempFile.get()
        iris.save(cube, temp)
        self.zonal_mean.set_local_file(temp, diagnostic=self, rename_var='result', region=self.basin)

    def _try_load_cube(self, number):
        try:
            cube = iris.load_cube('mesh_hgr.nc', 'e{0}{1}'.format(number, self.grid_point))
        except iris.exceptions.ConstraintMismatchError:
            cube = iris.load_cube('mesh_hgr.nc', 'e{0}{1}_0'.format(number, self.grid_point))
        cube = iris.util.squeeze(cube)
        dims = len(cube.shape)
        try:
            cube.coord('i')
        except iris.exceptions.CoordinateNotFoundError:
            cube.add_dim_coord(iris.coords.DimCoord(np.arange(cube.shape[dims - 1]), var_name='i'), dims - 1)
        try:
            cube.coord('j')
        except iris.exceptions.CoordinateNotFoundError:
            cube.add_dim_coord(iris.coords.DimCoord(np.arange(cube.shape[dims - 2]), var_name='j'), dims -2)
        return cube

    def _load_data(self):
        coords = []
        handler = Utils.open_cdf(self.variable_file.local_file)
        for variable in handler.variables:
            if variable in ('time', 'lev', 'lat', 'lon', 'latitude', 'longitude', 'leadtime', 'time_centered'):
                coords.append(variable)
            if variable == 'time_centered':
                handler.variables[variable].standard_name = ''

        handler.variables[self.variable].coordinates = ' '.join(coords)
        handler.close()

        data = iris.load_cube(self.variable_file.local_file)
        return self._rename_depth(data)

    def _rename_depth(self, data):
        for coord_name in ('model_level_number', 'Vertical T levels', 'lev'):
            if data.coords(coord_name):
                coord = data.coord(coord_name)
                coord.standard_name = 'depth'
                coord.long_name = 'depth'
                break
        return data

    def _fix_file_metadata(self):
        handler = Utils.open_cdf(self.variable_file.local_file)
        var = handler.variables[self.variable]
        coordinates = ''
        has_levels = False
        for dimension in handler.variables.keys():
            if dimension in ['time', 'lev', 'lat', 'latitude', 'lon', 'longitude', 'i', 'j']:
                coordinates += ' {0}'.format(dimension)
            if dimension == 'lev':
                has_levels = True
        var.coordinates = coordinates
        handler.close()
        return has_levels
